#!/usr/bin/env python3

import os.path as osp
import os
import time
import re
import json
import subprocess
from boaapi.boa_client import BoaClient, BOA_API_ENDPOINT
from boaapi.status import CompilerStatus, ExecutionStatus

def get_credentials():
    with open("credentials.txt", "r") as fh:
        creds = [ line.strip() for line in fh.readlines() ]
        user = creds[0]
        password = creds[1]
    fh.close()
    return (user, password)

def get_query_source(filename):
    fn_full = osp.join("boa-queries", filename)
    with open(fn_full, "r") as fh:
        data = fh.read()
    fh.close()
    return data

def replace_with_file(original, needle, filename):
    if re.search(needle, original) is not None:
        return replace_with_string(original, needle, get_query_source(filename))
    return original

def replace_with_string(original, needle, replacement):
    string = re.sub(needle, replacement, original)
    return string

def do_replacements(original, replacements):
    query = original
    for (needle, replace_with, is_file) in replacements:
        if is_file:
            query = replace_with_file(query, needle, replace_with)
        else:
            query = replace_with_string(query, needle, replace_with)
    return query

class BoaQuery:
    __datasets__ = {
        'kotlin': '2021 Aug/Kotlin',
        'java': '[admin] 2021 Aug/Java'
    }

    def __init__(self, name, file, dataset, user, password):
        self.name = name
        self.file = file
        self.dataset = dataset
        self.user = user
        self.password = password
        self.client = None

    def post_process(self):
        pass

    def prepare_query(self):
        pass

    def run(self):
        if self.client == None:
            self.client = BoaClient(endpoint = BOA_API_ENDPOINT)
            self.client.login(self.user, self.password)
        if not osp.exists(self.file):
            job = self.client.query(self.prepare_query(), self.client.get_dataset(self.__datasets__[self.dataset]))
            while job.is_running():
                job.refresh()
                print(f"Job {job.id} is running, waiting 10s.", flush=True)
                time.sleep(10)
            print(f"Job {job.id} is complete.")
            if job.compiler_status is CompilerStatus.ERROR:
                print(f"Job {job.id} ({self.name}) had a compilation error.", flush=True)
                print(job.get_compiler_errors(), flush=True)
            elif job.exec_status is ExecutionStatus.ERROR:
                print(f"Job {job.id} ({self.name}) had an execution error.", flush=True)
            else:
                job.set_public(True)
                base_path = osp.dirname(self.file)
                os.makedirs(base_path, exist_ok=True)
                time.sleep(180)
                url = job.get_public_url() + "/download"
                subprocess.run(["wget", "--quiet", "--compression=auto", "-O", self.file, url])
                name_parts = self.name.split("/")
                jobs_data = {}
                if osp.exists("jobs.json"):
                    with open("jobs.json", "r") as fh:
                        jobs_data = json.load(fh)
                    fh.close()
                cur_data = jobs_data
                for i, part in enumerate(name_parts):
                    if i == len(name_parts) - 1:
                        cur_data[part] = job.id
                    else:
                        if part not in cur_data.keys():
                            cur_data[part] = {}
                        cur_data = cur_data[part]
                with open("jobs.json", "w") as fh:
                    json.dump(jobs_data, fh,
                              indent=1)
                    fh.close()
                self.post_process()
