options(echo = FALSE)

source("scripts/common.R")

dataclasses_data <<- tibble()
classes_data <<- tibble()
interfaces_data <<- tibble()
num_parents_data <<- tibble()

read_data_files <- function(language, is_per_project) {
    
    if (language == "kotlin") {
        dataclasses_data <<- read_csv(sprintf("data-csv/kotlin-dataclasses%s.csv", if (is_per_project) {""} else {"-single"}),
                               col_names = c("type", "project", "file", "num_parents", "time_stamp", "commit", "count")) %>%
                select(-type)
    }
    classes_data <<- read_csv(sprintf("data-csv/%s-classes%s.csv", language, if(is_per_project) {""} else {"-single"}),
                       col_names = c("type", "project", "file", "is_open", "num_parents", "time_stamp", "commit", "count")) %>%
            select(-type)
    interfaces_data <<- read_csv(sprintf("data-csv/%s-interfaces%s.csv", language, if(is_per_project) {""} else {"-single"}),
                          col_names = c("type", "project", "file", "num_parents", "time_stamp", "commit", "count")) %>%
            select(-type)
    num_parents_data <<- read_csv(sprintf("data-csv/%s-num-parents%s.csv", language, if(is_per_project) {""} else {"-single"}),
                      col_names = c("type", "project", "path", "class", "timestamp", "revision", "num_parents")) %>%
        select(-type)
}

summarize_tables_and_join <- function(language, is_per_project) {
    data_classes_summary <- tibble()
    if (language == "kotlin") {
        data_classes_summary <- dataclasses_data %>% group_by(project)  %>% summarize(data_classes = sum(count))
    }
    classes_summary <- classes_data %>% group_by(project) %>% summarize(classes = sum(count))
    interfaces_summary <- interfaces_data %>% group_by(project) %>% summarize(interfaces = sum(count))
    complete_summary <- full_join(classes_summary, interfaces_summary)
    if (language == "kotlin") {
        complete_summary <- full_join(data_classes_summary, complete_summary)
    }
    complete_summary
}

build_summary_table <- function(summary_table, language, is_per_project) {
    pivot_cols <- if (language == "kotlin") {c("data_classes", "classes", "interfaces")} else {c("classes", "interfaces")}
    transposed_table <- summary_table %>%
        pivot_longer(cols = pivot_cols) %>%
	filter(!is.na(value)) %>%
	group_by(name) %>%
	summarize(count = sum(value),
    	          mean = mean(value),
		  std = sd(value),
		  min = min(value),
		  `25%` = quantile(value, probs=c(0.25)),
		  `50%` = quantile(value, probs=c(0.5)),
		  `75%` = quantile(value, probs=c(0.75)),
		  max = max(value)) %>%
	transpose_tibble(col_names = "name", id_col = " ") %>%
	rename("Classes" = classes, "Interfaces" = interfaces)
	
    if(language == "kotlin") {
        transposed_table <- transposed_table %>% rename("Data Classes" = data_classes)
    }

    transposed_table %>%
	xtable() %>% print(file = sprintf("tables/rq1-summary-%s-%s.tex", language, if (is_per_project) {"by-project"} else {"bag-of-classes"}))
    transposed_table
}

build_graph <- function(summary_table, language, is_per_project) {
    pivot_cols <- if (language == "kotlin") {c("data_classes", "classes", "interfaces")} else {c("classes", "interfaces")}
    summary_table %>%
        pivot_longer(cols = pivot_cols) %>%
	ggplot() +
	geom_boxplot2(aes(name, value), outlier.shape=NA) +
	theme_clean() +
	scale_color_colorblind()
    ggsave(sprintf("figures/rq1-summary-%s-%s.png", language, if (is_per_project) {"by-project"} else {"bag-of-classes"}))
}

read_data_files("kotlin", TRUE)
classes_kotlin <- classes_data %>% mutate(Language = "Kotlin")
num_parents_kotlin <- num_parents_data %>% mutate(Language = "Kotlin")

join_table_kotlin <- summarize_tables_and_join("kotlin", TRUE) %>% mutate(Language = "Kotlin")

transposed_join_table_kotlin <- build_summary_table(join_table_kotlin, "kotlin", TRUE)
transposed_join_table_kotlin <- transposed_join_table_kotlin %>% mutate(Language = "Kotlin")

build_graph(join_table_kotlin, "kotlin", TRUE)

read_data_files("java", TRUE)
classes_java <- classes_data %>% mutate(Language = "Java")
num_parents_java <- num_parents_data %>% mutate(Language = "Java")

join_table_java <- summarize_tables_and_join("java", TRUE) %>% mutate(Language = "Java")

transposed_join_table_java <- build_summary_table(join_table_java, "java", TRUE)
transposed_join_table_java <- transposed_join_table_java %>% mutate(Language = "Java")

build_graph(join_table_java, "java", TRUE)

full_join(transposed_join_table_kotlin, transposed_join_table_java) %>%
    pivot_wider(names_from = Language,
                values_from = c(Classes, `Data Classes`, Interfaces),
		names_glue = "{.value} ({Language})") %>%
    select(` `,
    	   `Classes (Java)`,
           `Classes (Kotlin)`,
	   `Data Classes (Kotlin)`,
   	   `Interfaces (Java)`,
	   `Interfaces (Kotlin)`) %>%
    xtable() %>% print(file = "tables/rq1-summary-compare.tex")

full_join(join_table_kotlin, join_table_java) %>%
    pivot_longer(cols = c("classes", "data_classes", "interfaces")) %>%
    mutate(name = recode_factor(name,
                                "classes" = "Classes",
				"data_classes" = "Data Classes", "interfaces" = "Interfaces")) %>%
    mutate(value = ifelse(is.na(value), 0, value)) %>%
    ggplot() +
    geom_boxplot2(aes(name, value, fill=Language), show.errorbar = TRUE, width.errorbar = 0.3) +
    theme_clean() +
    scale_color_colorblind() +
    ylab("Number per Project") +
    xlab("")
ggsave("figures/rq1-summary-compare.png", units="in", width=7, height=3.5)

full_join(classes_kotlin, classes_java) %>%
    mutate(is_open = ifelse(is_open, "Open", "Not Open")) %>%
    mutate(is_open = ifelse(is.na(is_open), "Open", is_open)) %>%
    group_by(Language, is_open, project) %>%
    summarize(n = sum(count)) %>%
    summarize(count = sum(n, na.rm = TRUE),
              mean = mean(n, na.rm = TRUE),
	      std = sd(n, na.rm = TRUE),
	      min = min(n, na.rm = TRUE),
	      `25%` = quantile(n, probs=c(0.25), na.rm = TRUE),
	      `50%` = quantile(n, probs=c(0.5), na.rm = TRUE),
	      `75%` = quantile(n, probs=c(0.75), na.rm = TRUE),
	      max = max(n, na.rm = TRUE)) %>%
    mutate(name = sprintf("%s (%s)", is_open, Language)) %>%
    ungroup() %>%
    select(-Language, -is_open) %>%
    transpose_tibble(col_names = "name", id_col = " ") %>%
    select(` `, `Open (Java)`, `Not Open (Java)`, `Open (Kotlin)`, `Not Open (Kotlin)`) %>%
    xtable() %>% print(file = "tables/rq1-open-closed-compare.tex")

full_join(num_parents_kotlin, num_parents_java) %>%
    filter(!is.na(num_parents)) %>%
    group_by(Language) %>%
    summarize(mean = mean(num_parents),
	      std = sd(num_parents),
	      min = min(num_parents),
	      `25%` = quantile(num_parents, probs=c(0.25)),
	      `50%` = quantile(num_parents, probs=c(0.5)),
	      `75%` = quantile(num_parents, probs=c(0.75)),
	      max = max(num_parents)) %>%
    transpose_tibble(col_names = "Language", id_col = " ") %>%
    xtable() %>% print(file = "tables/rq1-num-parents.tex")

full_join(num_parents_kotlin, num_parents_java) %>%
    filter(!is.na(num_parents)) %>%
    ggplot() +
    geom_boxplot2(aes(Language, num_parents), show.errorbar = TRUE, width.errorbar = 0.3) +
    theme_clean() +
    scale_color_colorblind() +
    ylab("Number of Parents") +
    xlab("")
ggsave("figures/rq1-num-parents.png", units="in", width=3, height = 3)

quit()