#!/usr/bin/env python3
# coding: utf-8

#imports
import time
import numpy as np
import pandas as pd
import os.path as osp
#from survival import *
from utils import *
from survival import run_survival_analysis
from scipy.stats import chi2_contingency, mannwhitneyu

def _create_cont_table(df,cat1, cat2):
    #creates contigency table given two categorical columns
    return pd.pivot_table(df[df['type']=='class'], values='count',index =[cat1],
                          columns = [cat2], aggfunc = np.sum).values

def _pure_and_mixed_kotlin(kotlin):
    kotlin['lang'] = kotlin['file'].apply(lambda x:'mixed' if (x.split('.')[-1]== 'java') else 'pure')
    mixed  = kotlin[kotlin.lang == 'mixed'].project.unique()
    mixed_df = kotlin[kotlin.project.isin(mixed)]
    mixed_df['mixed_or_pure'] = 'mixed'
    pure_df= kotlin[~kotlin.project.isin(mixed)]
    pure_df['mixed_or_pure'] = 'pure'
    return pd.concat([pure_df, mixed_df])

def size_vs_class_openess(jv, kt):
    '''
    Calculates the pearson's chisquare test.
    H0: There is no relationship between size of the project and openness of classes
    H1: There is a significant relationship between size of project and openness of classes
    '''
    #create contigency table
    jv_cont_table = _create_cont_table(jv, 'partition', 'isopen')
    kt_cont_table = _create_cont_table(kt, 'partition', 'isopen')
    #path to output
    path = "tables/size_matter.tex"
    #perform chisquare test of independence on java data
    stat,p, dof, _ = chi2_contingency(jv_cont_table)
    concl = 'dependent' if p<=0.05 else 'independent'
    df1 = {'Language':['java'],'Degree of Freedom': [dof],'p-value': [p],'Conclusion': [concl]}

    #erform chisquare test of independence on kotlin data
    stat,p, dof, _ = chi2_contingency(kt_cont_table)
    concl = 'dependent' if p<=0.05 else 'independent'
    df2 = {'Language':['kotlin'],'Degree of Freedom': [dof],'p-value': [p],'Conclusion': [concl]}

    df = pd.concat([pd.DataFrame.from_dict(df1), pd.DataFrame.from_dict(df2)])
    save_table(df, 'tables/does_size_matter_to_openness.tex')
#RQ1&RQ4: openness and inheritance (java vs. kotlin || kotln vs. kotlin-java)
#grouped by size
def _run_mann(dist1, dist2, rq, partition, hypothesis, alternative = 'greater'):
    stat, pvalue, warn = 0,0,""
    if len(dist1) != 0 and len(dist2) != 0:
        stat, pvalue = mannwhitneyu(x = dist1, y = dist2, alternative = alternative)
    elif len(dist1) == 0 and len(dist2) !=0:
        warn = "First distribution is empty!"
    elif len(dist1) != 0 and len(dist2) == 0:
        warn = f"{header}: Second distribution is empty"
    else:
        warn = f"{header}: Both distributions are empty!"


    return pd.DataFrame.from_dict({
        'RQ': [rq],
        'Partition': [partition],
        'Hypothesis': [hypothesis],
        'p-value': [pvalue],
        'Reject': [ 'Yes' if pvalue < 0.05 else 'No' ]
    })

def rq1_stats(java, kotlin, dit_noc):
    #claim 1: more open classes in java than kotlin(wilcoxon--mann-whitney test) || kotlin  than kotlin-java
    #claim 2: more closed classes in kotlin than java (wilcoxon--mann-whitnes test)
    #claim 3: much deeper inheritance in java than kotlin(one tail ttest) || kotlin-java than kotlin
    output = []
    java_num_parents = load_num_parents('java')
    kotlin_num_parents = load_num_parents('kotlin')
    for sz in ['small', 'medium', 'large']:
        #calculate the average number of open classes per project
        java_open   = java[java.isopen==True].groupby(['partition','project'], as_index = False).mean()
        kotlin_open = kotlin[kotlin.isopen == True].groupby(['partition','project'], as_index = False ).mean()
        #calculate the average number of closed classes per project
        java_closed   = java[java.isopen==False].groupby(['partition','project'], as_index = False).mean()
        kotlin_closed = kotlin[kotlin.isopen == False].groupby(['partition','project'], as_index = False ).mean()
        #average noc per project per language and size
        dit_noc       = dit_noc.groupby(['language','partition','project'], as_index = False).mean()
        #extract mean distributions per language
        dist_jv_open   = java_open[java_open.partition==sz]['count']
        dist_jv_closed = java_closed[java_closed.partition == sz]['count']
        dist_kt_open   = kotlin_open[kotlin_open.partition == sz]['count']
        dist_kt_closed = kotlin_closed[kotlin_closed.partition == sz]['count']
        dit_noc_jv     = dit_noc[dit_noc.partition == sz]['depth']
        dit_noc_kt     = dit_noc[dit_noc.partition == sz]['depth']
        java_np = java_num_parents[java_num_parents.partition == sz]['numparents']
        kotlin_np = kotlin_num_parents[kotlin_num_parents.partition == sz]['numparents']
        #perform the Mann-Whitney U test
        rq1_open   = _run_mann(dist_jv_open, dist_kt_open, 'RQ1', sz, 'There are more open classes in Java than in Kotlin.')
        rq1_closed = _run_mann(dist_kt_closed, dist_jv_closed, 'RQ1', sz, 'There are more closed classes in Kotlin than in Java.')
        deep_inher = _run_mann(dit_noc_jv, dit_noc_kt, 'RQ1', sz, 'Inheritance in Java is deeper than in Kotlin.')
        num_par = _run_mann(java_np, kotlin_np, 'RQ1', sz, 'Java classes inherit from more parents than do Kotlin classes.')
        #save resultls
        res = pd.concat([rq1_open, rq1_closed, deep_inher, num_par])
        output.append(res)

    # save_table(pd.concat(output), 'tables/RQ1_stats.tex')
    return pd.concat(output)
#RQ2: openness overtime
#survival analysis at the class and method level (remove the outlier)
def rq2_stats():
    run_survival_analysis()

def rq4_stats(kotlin, dit_noc):
    #RQ4
    output = []
    for sz in [ 'small', 'medium', 'large']:

        df = _pure_and_mixed_kotlin(kotlin)
        df_open   = df[df.isopen == True].groupby(['partition','mixed_or_pure','project'], as_index = False).mean()
        df_closed = df[df.isopen == False].groupby(['partition','mixed_or_pure', 'project'], as_index = False).mean()
        # dit_noc   = dit_noc.groupby(['partition','project'], as_index = False).mean()
        #extract mean distributions ped language
        dist_pr_open = df_open[(df_open.mixed_or_pure == 'pure') &(df_open.partition == sz)]['count']
        dist_mx_open = df_open[(df_open.mixed_or_pure == 'mixed') & (df_open.partition == sz)]['count']
        dist_pr_closed = df_closed[(df_closed.mixed_or_pure == 'pure') &(df_closed.partition == sz)]['count']
        dist_mx_closed = df_closed[(df_closed.mixed_or_pure == 'mixed') & (df_closed.partition == sz)]['count']
        # dit_noc_jv     = dit_noc[dit_noc.partition == sz]['depth']
        # dit_noc_kt     = dit_noc[dit_noc.partition == sz]['depth']
        #perform the Mann-Whitney U test
        rq4_open   = _run_mann(dist_mx_open, dist_pr_open, 'RQ4', sz, 'Kotlin-primary but mixed projects have more open classes that pure Kotlin projects.')
        rq4_closed = _run_mann(dist_pr_closed, dist_mx_closed, 'RQ4', sz,  'Pure Kotlin projects have more closed classes that Kotlin-primary but mixed projects.')
        # deep_inher = _run_mann(dit_noc_jv, dit_noc_kt, 'RQ4', size, 'RQ1_deeper_inheritance_java_vs_kotlin')
        #save resultls
        res = pd.concat([rq4_open,rq4_closed])
        output.append(res)

    return pd.concat(output)

def rq5():
    df_java = load_with_tests('java').groupby(['partition', 'project', 'istest', 'isopen'], as_index = False).count()
    df_kotlin = load_with_tests('kotlin').groupby(['partition', 'project', 'istest', 'isopen'], as_index = False).count()
    # df_both = pd.concat([df_java, df_kotlin])
    output = []
    for size in ['small', 'medium', 'large']:
        dist_kt_test_open = df_kotlin[(df_kotlin.istest == True) & (df_kotlin.isopen == True)]['count']
        dist_kt_test_closed = df_kotlin[(df_kotlin.istest == True) & (df_kotlin.isopen == False)]['count']
        dist_jv_test_open = df_java[(df_java.istest == True) & (df_java.isopen == True)]['count']
        dist_jv_test_closed = df_java[(df_java.istest == True) & (df_java.isopen == False)]['count']
        kotlin_open = _run_mann(dist_kt_test_open, dist_kt_test_closed, 'RQ5', size, 'Kotlin test classes are more likely to be open than closed.')
        java_open = _run_mann(dist_jv_test_open, dist_jv_test_closed, 'RQ5', size, 'Java test classes are more likely to be open than closed.')
        res = pd.concat([kotlin_open, java_open])
        output.append(res)

    return pd.concat(output)

def rq3():
    df_java = load_class_churn_data('java')
    df_kotlin = load_class_churn_data('kotlin')
    df_java_open = df_java[df_java['isopen'] == True]
    df_java_closed = df_java[df_java['isopen'] == False]
    df_kotlin_open = df_kotlin[df_kotlin['isopen'] == True]
    df_kotlin_closed = df_kotlin[df_kotlin['isopen'] == False]
    output = []
    for size in ['small', 'medium', 'large']:
        dist_java_open = df_java_open[df_java_open['partition'] == size]['churn_rate']
        dist_java_closed = df_java_closed[df_java_closed['partition'] == size]['churn_rate']
        dist_kotlin_open = df_kotlin_open[df_kotlin_open['partition'] == size]['churn_rate']
        dist_kotlin_closed = df_kotlin_closed[df_kotlin_closed['partition'] == size]['churn_rate']
        kotlin_open = _run_mann(dist_kotlin_open, dist_kotlin_closed, 'RQ3', size, 'Open classes in Kotlin have higher rates of churn that closed classes.')
        java_open = _run_mann(dist_java_open, dist_java_closed, 'RQ3', size, 'Open classes in Java have higher rates of churn that closed classes.')
        res = pd.concat([kotlin_open, java_open])
        output.append(res)

    return pd.concat(output)
    # save_table(pd.concat(output), 'tables/rq4-tests.tex')

def java_kotlin_stats(java, kotlin, ditnoc):
    size_vs_class_openess(java, kotlin)
    out = []
    out.append(rq1_stats(java, kotlin, ditnoc))
    rq2_stats()
    out.append(rq3())
    out.append(rq4_stats(kotlin, ditnoc))
    out.append(rq5())
    df = pd.concat(out)
    df['partnum'] = df['Partition'].apply(lambda x: {'small': 1, 'medium': 2, 'large': 3}[x])
    df = df.sort_values(['RQ','Hypothesis', 'partnum'])
    del df['partnum']
    save_table(df, 'tables/hyp-testing.tex', index_names = False, multicolumn_format = 'c', multirow = True, index = False)

def percent_open_per_project(frame1,frame2):
    '''
    @frame1: java dataframe with project:file:classname headers
    @frame2: kotlin dataframe with project:file:classname headers
    @return: a tuple of length 2 containing two arrays. First one is an array of open:close ratio
    per project in java, and similarly another array for kotlin
    '''
    frame1 = frame1[frame1['file'] != 'DUMMYDUMMYDUMMY.java']
    frame2 = frame2[frame2['file'] != 'DUMMYDUMMYDUMMY.java']
    df1 = pd.crosstab(frame1.project, frame1.isopen).reset_index()
    df2 = pd.crosstab(frame2.project, frame2.isopen).reset_index()
    try:
        df1['percent_open'] = (df1[True]/(df1[True] + df1[False])) * 100
    except KeyError:
        try:
            df1['percent_open'] = (df1[True]/df1[True]) * 100 # no closed classes
        except KeyError:
            df1['percent_open'] = 0 # no open classes
    try:
        df2['percent_open'] = (df2[True]/(df2[True] + df2[False])) * 100
    except KeyError:
        try:
            df2['percent_open'] = (df2[True]/df2[True]) * 100 #no closed classes
        except KeyError:
            df2['percent_open'] = 0 #no open classes

    return (df1.percent_open, df2.percent_open)

def split_kotlin(kt):
    '''
    @kt: kotlin.csv data file
    #return: tuple(pure_kotlin_project, mixed_kotlin_project)
    '''
    ktt = kt.copy()
    ktt = ktt[ktt['file'] != "DUMMYDUMMYDUMMY.java"]
    ktt['lang'] = ktt['file'].apply(lambda x: x.split('.')[-1])
    mixed_projects = list(ktt[ktt.lang =='java'].project)
    return(kt[~kt.project.isin(mixed_projects)],kt[kt.project.isin(mixed_projects)] )
    
def main():
    print('data loading ...')
    t0 = time.time()
    java = load_joined('java')
    kotlin = load_joined('kotlin')
    ditnoc = load_dit_noc()
    t1 = time.time()
    print('running statistical test ...')
    java_kotlin_stats(java, kotlin, ditnoc)
    t2 = time.time()
    print(f"Time taken:\ndata loading...{(t1-t0)/60}min\nrunning test...{(t2-t1)/60}min")
    print('done.')
#RQ3: change frequency (Sam)
#RQ5: testing code

if __name__ == '__main__':
    main()
