#!/usr/bin/env python
# coding: utf-8

import pandas as pd
import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from scipy.stats import zscore
import os.path as osp
import seaborn as sns
import matplotlib.pyplot as plt
import re
from cluster import partition

# Set default plot style
sns.set_style('whitegrid')
sns.set_palette('colorblind')

def save_figure(figure, filename, x=None, y=None):
    '''Save a FIGURE to FILENAME with size of X, Y inches.'''
    fig = figure.get_figure()
    plt.tight_layout()
    if x is not None:
        fig.set(figwidth = x)
    if y is not None:
        fig.set(figheight = y)
    fig.savefig(filename, dpi=600)
    plt.close(fig)

def make_figure(data, x, y, hue, condition, base):
    '''Make a comparison figure from DATA (appropriately grouped and
summarized), of size X, Y inches.  Filenames are form BASE-true.png,
BASE-false.png, where true and false refer to the CONDITION, i.e., a
series describing included row-keys.
    '''
    data_true = data[condition]
    data_false = data[condition.apply(lambda x:not x)]
    save_figure(sns.boxplot(x=x, y=y, hue=hue, data=data_true), f"{base}-true.png")
    save_figure(sns.boxplot(x=x, y=y, hue=hue, data=data_false), f"{base}-false.png")

def make_summary_table(data, group_by, keep, filename, type='sum', highlight_max=False):
    df = data.groupby(group_by)
    df = df[keep]
    if type == 'sum':
        df = df.sum().describe()
    elif type == 'count':
        df = df.count().describe()
    elif type == 'mean':
        df = df.mean().describe()
    else:
        df.describe()
    if highlight_max:
        s = df.style.highlight_max(axis=None, props='bfseries: ;')
        save_table(s, filename)
    else:
        save_table(df, filename, index_names=False, multirow=True)

colsepname = ''
def save_table(df, filename, decimals=2, colsep=False, thousands=',', **kwargs):
    global colsepname
    if not colsep is False:
        colsepname = colsepname + 'A'

    pd.set_option('styler.latex.hrules', True)
    pd.set_option('styler.latex.multicol_align', 'c')
    pd.options.display.float_format = ('{:,.' + str(decimals) + 'f}').format

    with pd.option_context("max_colwidth", 1000):
        styler = df.style
        styler = styler.applymap_index(lambda x: 'textbf:--rwrap;', axis='index')
        styler = styler.hide(names=True, axis='index')
        styler = styler.applymap_index(lambda x: 'textbf:--rwrap;', axis='columns')
        styler = styler.hide(names=True, axis='columns')
        tab1 = styler.format(None,
                             precision=decimals,
                             thousands=thousands,
                             escape='latex').to_latex(**kwargs)
    # print(tab1)
    with open(filename,'w',encoding='utf-8') as f:
        f.write('% DO NOT EDIT\n')
        f.write('% this file was automatically generated\n')
        if not colsep is False:
            f.write('\\newcommand{\\oldtabcolsep' + colsepname + '}{\\tabcolsep}\n')
            f.write('\\renewcommand{\\tabcolsep}{' + colsep + '}\n')
        f.write(tab1)
        if not colsep is False:
            f.write('\\renewcommand{\\tabcolsep}{\\oldtabcolsep' + colsepname +'}\n')


def load_data_csv(language, file, **kwargs):
    return pd.read_csv(f"data-csv/{language}/{file}", on_bad_lines='skip', **kwargs)

def load_exclude_projects(language):
    path = f"data-csv/cached/{language}-exclude-projects.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    else:
        frames = [load_data_csv(language, "dedupe-classes.csv")[["project", "file"]], load_data_csv(language, "dedupe-interfaces.csv")[["project", "file"]]]
        if language == 'kotlin':
            frames.append(load_data_csv(language, 'dedupe-data-classes.csv')[["project", "file"]])
        df = pd.concat(frames)
        df = df[['project', 'file']].groupby('project', as_index = False).count()
        df = df.set_index('project').apply(zscore).reset_index()
        df = df[abs(df['file']) >= 3]
        df = df[['project']]
        df.to_parquet(path)
        return df

def partition(data):
    '''
    Data: A dataframe guaranteed to have [project, file] columns
    Based on the number of files, we calculate zscore
    Large projects are those with zscore >=3
    Small project are below the median but below zscore=3
    Medium project are above the median but below zscore=3
    '''
    df = data[['project','file']].groupby('project', as_index = False).count()
    df_ = df.set_index('project').apply(zscore).reset_index()

    lrg = df_[df_['file']>=3] #large projects
    lrg['partition'] = 'large'

    rest= df[df.project.isin(df_[df_['file']<3].project)]
    sml = df[df['file']< np.median(rest['file'])] #small projects
    sml['partition'] = 'small'
    med = df[df['file'] >= np.median(rest['file'])] #medium project
    med['partition'] = 'medium'

    return pd.concat([sml[['project','partition']],med[['project','partition']], lrg[['project','partition']]])

def load_classes(language):
    path = f"data-csv/cached/{language}-classes.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)

    raw = f"data-csv/{language}/dedupe-classes.csv"
    df = pd.read_csv(raw, on_bad_lines = 'skip')
    partitions = load_partitions(language)
    df = df.merge(partitions, on=['project'], how = 'left')
    df.to_parquet(path, partition_cols = ['partition'])
    return df

def load_partitions(language):
    path = f"data-csv/cached/{language}-partitions.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    frames = [load_data_csv(language, "dedupe-classes.csv")[["project", "file"]], load_data_csv(language, "dedupe-interfaces.csv")[["project", "file"]]]
    if language == 'kotlin':
        frames.append(load_data_csv(language, 'dedupe-data-classes.csv')[["project", "file"]])
    df = pd.concat(frames)
    df = partition(df)
    df.to_parquet(path)
    return df

def load_num_parents(language):
    path = f"data-csv/cached/{language}-num-parents.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    df_excludes = load_exclude_projects(language)
    df_num_parents = load_data_csv(language, 'dedupe-num-parents.csv')
    df_merged = df_num_parents.merge(df_excludes, how='outer', indicator=True)
    df_filtered = df_merged[df_merged['_merge'] == 'left_only'].drop('_merge', axis='columns')
    df_filtered.to_parquet(path)
    return df_filtered

def load_joined(language):
    path = f"data-csv/cached/{language}.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    df_partitions = load_partitions(language)
    frames = []
    df_classes = load_data_csv(language, "dedupe-classes.csv")
    df_classes['type'] = 'class'
    df_classes = df_classes.merge(df_partitions, on=['project'], how='left')
    frames.append(df_classes)
    df_interfaces = load_data_csv(language, 'dedupe-interfaces.csv')
    df_interfaces['isopen'] = True
    df_interfaces['type'] = 'interface'
    df_interfaces = df_interfaces.merge(df_partitions, on=['project'], how='left')
    frames.append(df_interfaces)
    if language == 'kotlin':
        df_dataclasses = load_data_csv(language, 'dedupe-data-classes.csv')
        df_dataclasses['type'] = 'dataclass'
        df_dataclasses['isopen'] = False
        df_dataclasses.merge(df_partitions, on=['project'], how = 'left')
        frames.append(df_dataclasses)
    df_joined = pd.concat(frames, ignore_index = True)
    df_joined['language'] = language
    df_excludes = load_exclude_projects(language)
    df_merged = df_joined.merge(df_excludes, how='outer', indicator=True)
    print(df_merged.columns)
    df_filtered = df_merged[df_merged['_merge'] == 'left_only'].drop('_merge', axis='columns')
    df_filtered.to_parquet(path)
    return df_filtered

def load_class_interface_data():
    if osp.exists("data-csv/cached/classes-interfaces.parquet"):
        return pd.read_parquet("data-csv/cached/classes-interfaces.parquet")
    else:
        df_joined = pd.concat([load_joined('java'), load_joined('kotlin')], ignore_index = True)
        df_joined.to_parquet("data-csv/cached/classes-interfaces.parquet")
        return df_joined

def load_project_counts():
    if osp.exists("data-csv/cached/project-counts.parquet"):
        return pd.read_parquet("data-csv/cached/project-counts.parquet")
    else:
        df = load_class_interface_data()
        df_normal = df.groupby(['language', 'project', 'type'], as_index=False).sum('count').drop(['isopen', 'numparents', 'timestamp'], axis = 1)
        df_files = df.drop_duplicates(subset=['language', 'project', 'file']).groupby(['language', 'project'], as_index = False).count()[['language', 'project', 'count']]
        df_files['type'] = 'file'
        df_joined = pd.concat([df_normal, df_files], ignore_index = True)
        df_pivot = pd.pivot_table(df_joined, index=['language', 'project'], columns='type', values='count', fill_value = 0).reset_index()
        df_pivot.to_parquet("data-csv/cached/project-counts.parquet")
        return df_pivot

def load_tests_data(language):
    if osp.exists(f"data-csv/cached/tests-{language}.parquet"):
        return pd.read_parquet(f"data-csv/cached/tests-{language}.parquet")
    else:
        df_tests = load_data_csv(language, "dedupe-tests.csv")
        del df_tests['type']
        df_tests['istest'] = True
        df_tests.to_parquet(f"data-csv/cached/tests-{language}.parquet")
        return df_tests

def load_with_tests(language):
    if osp.exists(f"data-csv/cached/{language}-with-tests.parquet"):
        return pd.read_parquet(f"data-csv/cached/{language}-with-tests.parquet")
    else:
        df_tests = load_tests_data(language)
        df_data = load_joined(language)
        df_joined = df_data.merge(df_tests, on=['project', 'file'], how='left')
        df_joined['istest'] = df_joined['istest'].apply(lambda x: x == True)
        df_joined.to_parquet(f"data-csv/cached/{language}-with-tests.parquet")
        return df_joined

def load_all_tests_data():
    if osp.exists(f"data-csv/cached/all-tests.parquet"):
        return pd.read_parquet(f"data-csv/cached/all-tests.parquet")
    else:
        df_joined = pd.concat([load_with_tests('java'), load_with_tests('kotlin')])
        df_joined.to_parquet(f"data-csv/cached/all-tests.parquet")
        return df_joined

def load_dit_noc():
    if osp.exists("data-csv/cached/ditnoc.parquet"):
        return pd.read_parquet("data-csv/cached/ditnoc.parquet")
    else:
        df_java_dit_noc = load_data_csv("java", "dit-noc-processed.csv")
        df_java_dit_noc['language'] = 'Java'
        df_java_partitions = load_partitions('java')
        df_java = pd.merge(df_java_dit_noc, df_java_partitions, how = 'left', on = ['project'])
        df_kotlin_dit_noc = load_data_csv("kotlin", "dit-noc-processed.csv")
        df_kotlin_dit_noc['language'] = 'Kotlin'
        df_kotlin_partitions = load_partitions('kotlin')
        df_kotlin = pd.merge(df_kotlin_dit_noc, df_kotlin_partitions, how = 'left', on = ['project'])
        df_joined = pd.concat([df_java, df_kotlin], ignore_index = True)
        df_joined.to_parquet("data-csv/cached/ditnoc.parquet")
        return df_joined

def load_class_churn_data(language):
    path = f"data-csv/cached/churn-{language}.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    else:
        df_excludes = load_exclude_projects(language)
        df_commits = load_data_csv(language, 'churn-commits.csv')
        del df_commits['type']
        df_commits.rename(columns = {'count': 'total_commits'}, inplace = True)
        df_class = load_data_csv(language, "churn-class.csv", dtype={'type': str, 'project': 'Int64', 'file': str, 'isopen': bool, 'class': str, 'count': 'Int64'})
        df_class['type'] = 'class'
        df_class.rename(columns = {'is_open': 'isopen'}, inplace=True)
        df_class = pd.merge(df_class, df_commits, how = 'left', on = 'project')
        df_class['churn_rate'] = df_class.apply(lambda x: x['count']/x['total_commits'], axis=1)
        df_merged = df_class.merge(df_excludes, how='outer', indicator=True)
        df_filtered = df_merged[df_merged['_merge'] == 'left_only'].drop('_merge', axis='columns')
        df_filtered.to_parquet(path)
        return df_filtered

def load_method_churn_data(language):
    path = f"data-csv/cached/churn-method-{language}.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    else:
        df_partitions = load_partitions(language)
        df_commits = load_data_csv(language, 'churn-commits.csv')
        del df_commits['type']
        df_commits.rename(columns = {'count': 'total_commits'}, inplace = True)
        df_method = load_data_csv(language, 'churn-method.csv')
        df_method['type'] = 'method'
        df_method.rename(columns = {'is_open': 'isopen'}, inplace = True)
        df_method = pd.merge(df_method, df_partitions, how = 'left', on = 'project')
        df_method = pd.merge(df_method, df_commits, how = 'left', on = 'project')
        df_method['churn_rate'] = df_method.apply(lambda x: x['count']/x['total_commits'], axis=1)
        df_method.to_parquet(path)
        return df_method

def load_survival_data(lang, openness = True):
    path = f'data-csv/cached/{lang}-survival-data-open-to-close-is-{openness}.parquet'
    if osp.exists(path):
        return pd.read_parquet(path)
    raw =  f'data-csv/{lang}/dedupe-survival-classes.csv'
    df = pd.read_csv(raw, on_bad_lines = 'skip')
    df = df[df.startstatus == openness] #either open_to_close or close_to_open
    df = df[df.beforecollection >= 0] #filter those with negative times
    df['time'] = df.apply(lambda x: x['timetochange'] if x['timetochange']>=0 else x['beforecollection'], axis = 1)
    #df.rename(columns = {'startstatus': 'status', 'beforecollection': 'start'}, inplace = True)
    df['status'] = df['timetochange'].apply(lambda x: x > 0)
    df['time']   = df['time'].apply(lambda x: x/86400000000)
    df = df[['time','status']]
    df ['language'] = lang
    df.to_parquet(path)
    return df

def load_evolution_data(language):
    path = f'data-csv/cached/{language}-evolution.parquet'
    if osp.exists(path):
        return pd.read_parquet(path)
    df_excludes = load_exclude_projects(language)
    data = load_data_csv(language, "dedupe-evolution-classes.csv")
    df_merged = data.merge(df_excludes, how='outer', indicator=True)
    data = df_merged[df_merged['_merge'] == 'left_only'].drop('_merge', axis='columns')
    data['first_status'] = data['isopen']
    data['last_status'] = data['isopen']
    data['num_changes'] = data['timestamp']
    data_grouped = data.groupby(['project', 'file', 'class'],
                                as_index = False).aggregate(
                                    {'num_changes': lambda x: len(x) - 1,
                                     'first_status': lambda x: x.iloc[0],
                                     'last_status': lambda x: x.iloc[-1]})
    data_grouped.to_parquet(path)
    return data_grouped
    
def load_all_evolution_data():
    path = "data-csv/cached/all-evolution.parquet"
    if osp.exists(path):
        return pd.read_parquet(path)
    def categorize(x):
        if x['num_changes'] == 0:
            return 'Always Open' if x['first_status'] else 'Always Closed'
        elif x['num_changes'] == 1:
            return 'Becomes Open' if x['last_status'] else 'Becomes Closed'
        else:
            return "Multiple Changes"
    java = load_evolution_data('java')
    java['language'] = 'Java'
    kotlin = load_evolution_data('kotlin')
    kotlin['language'] = 'Kotlin'
    df = pd.concat([java, kotlin])
    df['pattern'] = df.apply(categorize, axis=1)
    df.to_parquet(path)
    return df
    
